def read_100_stitches(f, out):
    count = 0
    while True:
        count += 1
        b = bytearray(f.read(4))
        if len(b) != 4:
            break
        x = b[2]
        y = b[3]
        if x > 0x80:
            x -= 0x80
            x = -x
        if y > 0x80:  # because 2s complement is for chumps?
            y -= 0x80
            y = -y
        if b[0] == 0x61:
            out.stitch(x, -y)
            continue
        elif (b[0] & 0x01) != 0:
            out.move(x, -y)
            continue
        else:  # too broad of catch
            out.color_change()
            continue
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    read_100_stitches(f, out)



from __future__ import print_function

import unittest
from pattern_for_tests import *


class TestDataPreservation(unittest.TestCase):

    def test_preserve_u01_needles(self):
        file1 = "file.u01"
        pattern = get_big_pattern()
        pattern.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, None, 6, 0))
        write_u01(pattern, file1)
        read_pattern = read_u01(file1)
        for cmd in read_pattern.get_match_commands(NEEDLE_SET):
            print(decode_embroidery_command(cmd[2]))
        self.addCleanup(os.remove, file1)



from .EmbThread import EmbThread


def get_thread_set():
    return [
        None, #  EmbThreadPec(0, 0, 0, "Unknown", "0"),
        EmbThreadPec(14, 31, 124, "Prussian Blue", "1"),
        EmbThreadPec(10, 85, 163, "Blue", "2"),
        EmbThreadPec(0, 135, 119, "Teal Green", "3"),
        EmbThreadPec(75, 107, 175, "Cornflower Blue", "4"),
        EmbThreadPec(237, 23, 31, "Red", "5"),
        EmbThreadPec(209, 92, 0, "Reddish Brown", "6"),
        EmbThreadPec(145, 54, 151, "Magenta", "7"),
        EmbThreadPec(228, 154, 203, "Light Lilac", "8"),
        EmbThreadPec(145, 95, 172, "Lilac", "9"),
        EmbThreadPec(158, 214, 125, "Mint Green", "10"),
        EmbThreadPec(232, 169, 0, "Deep Gold", "11"),
        EmbThreadPec(254, 186, 53, "Orange", "12"),
        EmbThreadPec(255, 255, 0, "Yellow", "13"),
        EmbThreadPec(112, 188, 31, "Lime Green", "14"),
        EmbThreadPec(186, 152, 0, "Brass", "15"),
        EmbThreadPec(168, 168, 168, "Silver", "16"),
        EmbThreadPec(125, 111, 0, "Russet Brown", "17"),
        EmbThreadPec(255, 255, 179, "Cream Brown", "18"),
        EmbThreadPec(79, 85, 86, "Pewter", "19"),
        EmbThreadPec(0, 0, 0, "Black", "20"),
        EmbThreadPec(11, 61, 145, "Ultramarine", "21"),
        EmbThreadPec(119, 1, 118, "Royal Purple", "22"),
        EmbThreadPec(41, 49, 51, "Dark Gray", "23"),
        EmbThreadPec(42, 19, 1, "Dark Brown", "24"),
        EmbThreadPec(246, 74, 138, "Deep Rose", "25"),
        EmbThreadPec(178, 118, 36, "Light Brown", "26"),
        EmbThreadPec(252, 187, 197, "Salmon Pink", "27"),
        EmbThreadPec(254, 55, 15, "Vermillion", "28"),
        EmbThreadPec(240, 240, 240, "White", "29"),
        EmbThreadPec(106, 28, 138, "Violet", "30"),
        EmbThreadPec(168, 221, 196, "Seacrest", "31"),
        EmbThreadPec(37, 132, 187, "Sky Blue", "32"),
        EmbThreadPec(254, 179, 67, "Pumpkin", "33"),
        EmbThreadPec(255, 243, 107, "Cream Yellow", "34"),
        EmbThreadPec(208, 166, 96, "Khaki", "35"),
        EmbThreadPec(209, 84, 0, "Clay Brown", "36"),
        EmbThreadPec(102, 186, 73, "Leaf Green", "37"),
        EmbThreadPec(19, 74, 70, "Peacock Blue", "38"),
        EmbThreadPec(135, 135, 135, "Gray", "39"),
        EmbThreadPec(216, 204, 198, "Warm Gray", "40"),
        EmbThreadPec(67, 86, 7, "Dark Olive", "41"),
        EmbThreadPec(253, 217, 222, "Flesh Pink", "42"),
        EmbThreadPec(249, 147, 188, "Pink", "43"),
        EmbThreadPec(0, 56, 34, "Deep Green", "44"),
        EmbThreadPec(178, 175, 212, "Lavender", "45"),
        EmbThreadPec(104, 106, 176, "Wisteria Violet", "46"),
        EmbThreadPec(239, 227, 185, "Beige", "47"),
        EmbThreadPec(247, 56, 102, "Carmine", "48"),
        EmbThreadPec(181, 75, 100, "Amber Red", "49"),
        EmbThreadPec(19, 43, 26, "Olive Green", "50"),
        EmbThreadPec(199, 1, 86, "Dark Fuschia", "51"),
        EmbThreadPec(254, 158, 50, "Tangerine", "52"),
        EmbThreadPec(168, 222, 235, "Light Blue", "53"),
        EmbThreadPec(0, 103, 62, "Emerald Green", "54"),
        EmbThreadPec(78, 41, 144, "Purple", "55"),
        EmbThreadPec(47, 126, 32, "Moss Green", "56"),
        EmbThreadPec(255, 204, 204, "Flesh Pink", "57"),
        EmbThreadPec(255, 217, 17, "Harvest Gold", "58"),
        EmbThreadPec(9, 91, 166, "Electric Blue", "59"),
        EmbThreadPec(240, 249, 112, "Lemon Yellow", "60"),
        EmbThreadPec(227, 243, 91, "Fresh Green", "61"),
        EmbThreadPec(255, 153, 0, "Orange", "62"),
        EmbThreadPec(255, 240, 141, "Cream Yellow", "63"),
        EmbThreadPec(255, 200, 200, "Applique", "64")
    ]


class EmbThreadPec(EmbThread):
    def __init__(self, red, green, blue, description, catalog_number):
        EmbThread.__init__(self)
        self.set_color(red, green, blue)
        self.description = description
        self.catalog_number = catalog_number
        self.brand = "Brother"
        self.chart = "Brother"



from pyembroidery import *

import math


def evaluate_lsystem(symbol, rules, depth):
    if depth <= 0 or symbol not in rules:
        symbol()
    else:
        for produced_symbol in rules[symbol]:
            evaluate_lsystem(produced_symbol, rules, depth - 1)


class Turtle:
    def __init__(self, pattern):
        self.pattern = pattern
        self.angle = 0
        self.x = 0
        self.y = 0
        import math
        self.turn_amount = math.pi / 3

    def forward(self, distance):
        self.x += distance * math.cos(self.angle)
        self.y += distance * math.sin(self.angle)
        self.pattern.add_stitch_absolute(STITCH, self.x, self.y)
        # self.pattern.add_stitch_absolute(SEQUIN_EJECT, self.x, self.y)

    def turn(self, angle):
        self.angle += angle

    def move(self, distance):
        self.x += distance * math.cos(self.angle)
        self.y += distance * math.sin(self.angle)

    def add_gosper(self):
        a = lambda: self.forward(20)
        b = lambda: self.forward(20)
        l = lambda: self.turn(self.turn_amount)
        r = lambda: self.turn(-self.turn_amount)
        initial = lambda: None
        rules = {
            initial: [a],
            a: [a, l, b, l, l, b, r, a, r, r, a, a, r, b, l],
            b: [r, a, l, b, b, l, l, b, l, a, r, r, a, r, b]
        }
        evaluate_lsystem(initial, rules, 3)  # 4

    def add_serp(self):
        a = lambda: self.forward(20)
        b = lambda: self.forward(20)
        l = lambda: self.turn(self.turn_amount)
        r = lambda: self.turn(-self.turn_amount)
        initial = lambda: None
        rules = {
            initial: [a],
            a: [b, l, a, l, b],
            b: [a, r, b, r, a]
        }
        evaluate_lsystem(initial, rules, 3)  # 6

def get_big_pattern():
    pattern = EmbPattern()
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "red")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "blue")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "green")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "grey")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "gold")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "ivory")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "khaki")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "oldlace")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "olive")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "pink")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "purple")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "tan")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "violet")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "white")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "salmon")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "lime")  # 16 blocks.
    return pattern


def get_shift_pattern():
    pattern = EmbPattern()
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "red")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "blue")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "green")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "grey")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "gold")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "ivory")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "khaki")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "oldlace")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "olive")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "pink")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "purple")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "tan")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "violet")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "white")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "salmon")
    pattern.add_command(MATRIX_TRANSLATE, 25, 25)
    pattern.add_command(MATRIX_ROTATE, 22.5)
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "lime")
    return pattern


def get_simple_pattern():
    pattern = EmbPattern()
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "red")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "blue")
    pattern.add_block([(0, 0), (0, 100), (100, 100), (100, 0), (0, 0)], "green")
    return pattern


def get_random_pattern_large(count=1000):
    pattern = EmbPattern()
    import random

    for i in range(0, count):
        pattern.add_block(
            [(random.uniform(-500, 500), random.uniform(-500, 500)),
             (random.uniform(-500, 500), random.uniform(-500, 500)),
             (random.uniform(-500, 500), random.uniform(-500, 500))],
            random.randint(0x000000, 0xFFFFFF))
    return pattern


def get_random_pattern_small():
    pattern = EmbPattern()
    import random

    pattern.add_block(
        [(random.uniform(-500, 500), random.uniform(-500, 500)),
         (random.uniform(-500, 500), random.uniform(-500, 500)),
         (random.uniform(-500, 500), random.uniform(-500, 500))],
        random.randint(0x000000, 0xFFFFFF))
    return pattern


def get_random_pattern_small_halfs():
    pattern = EmbPattern()
    import random
    pattern.add_block(
        [(random.randint(-500, 500) / 2.0, random.randint(-500, 500) / 2.0),
         (random.randint(-500, 500) / 2.0, random.randint(-500, 500) / 2.0),
         (random.randint(-500, 500) / 2.0, random.randint(-500, 500) / 2.0)],
        random.randint(0x000000, 0xFFFFFF))
    return pattern


def get_fractal_pattern():
    pattern = EmbPattern()
    turtle = Turtle(pattern)
    turtle.add_gosper()
    pattern.add_command(COLOR_BREAK)
    turtle.move(500)
    turtle.add_serp()
    pattern.add_command(SEQUENCE_BREAK)
    pattern.add_command(STOP)
    turtle.move(50)
    turtle.add_serp()
    pattern.add_command(SEQUENCE_BREAK)
    turtle.turn(-math.pi / 3)
    turtle.move(500)
    turtle.add_serp()
    pattern.add_command(COLOR_BREAK)
    turtle.turn(-math.pi / 3)
    turtle.move(500)  # 260, -450
    turtle.add_gosper()
    pattern.add_command(END)
    return pattern



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestEmbpattern(unittest.TestCase):

    def test_encoder_bookend_color_break(self):
        pattern = EmbPattern()
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("red")
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern = pattern.get_normalized_pattern()
        self.assertEqual(len(pattern.threadlist), 1)
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE), 0)

    def test_encoder_multiple_internal_breaks(self):
        pattern = EmbPattern()
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("red")
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("green")
        pattern.add_command(COLOR_BREAK)
        pattern = pattern.get_normalized_pattern()
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE), 4)
        self.assertEqual(len(pattern.threadlist), 5)

    def test_encoder_colorchange(self):
        pattern = EmbPattern()
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("red")
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("green")
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("blue")
        pattern.add_command(COLOR_BREAK)
        pattern = pattern.get_normalized_pattern()
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE) + 1, len(pattern.threadlist))

    def test_encoder_needleset(self):
        pattern = EmbPattern()
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("red")
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("green")
        pattern.add_command(COLOR_BREAK)
        pattern.stitch_abs(0, 0)
        pattern.stitch_abs(0, 100)
        pattern.stitch_abs(100, 100)
        pattern.stitch_abs(100, 0)
        pattern.stitch_abs(0, 0)
        pattern.add_thread("blue")
        pattern.add_command(COLOR_BREAK)
        pattern = pattern.get_normalized_pattern({"thread_change_command": NEEDLE_SET})
        self.assertEqual(pattern.count_stitch_commands(NEEDLE_SET), len(pattern.threadlist))

    def test_transcode_to_self(self):
        pattern = get_shift_pattern()
        from pyembroidery.EmbEncoder import Transcoder
        encoder = Transcoder()
        encoder.transcode(pattern, pattern)
        self.assertNotEquals(len(pattern.stitches), 0)



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestConverts(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_convert_exp_to_u01(self):
        file1 = "convert_u01.exp"
        file2 = "converted_exp.u01"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_u01(f_pattern, file2)
        t_pattern = read_u01(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(NEEDLE_SET), 16)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->u01: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_csv(self):
        file1 = "convert_csv.exp"
        file2 = "converted_exp.csv"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_csv(f_pattern, file2)
        t_pattern = read_csv(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->csv: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_exp(self):
        file1 = "convert_exp.exp"
        file2 = "converted_exp.exp"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_exp(f_pattern, file2)
        t_pattern = read_exp(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->exp: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_pes(self):
        file1 = "convert_pes.exp"
        file2 = "converted_exp.pes"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_pes(f_pattern, file2)
        t_pattern = read_pes(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->pes: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_jef(self):
        file1 = "convert_jef.exp"
        file2 = "converted_exp.jef"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_jef(f_pattern, file2)
        t_pattern = read_jef(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->jef: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_pec(self):
        file1 = "convert_pec.exp"
        file2 = "converted_exp.pec"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_pec(f_pattern, file2)
        t_pattern = read_pec(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->pec: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_vp3(self):
        file1 = "convert_vp3.exp"
        file2 = "converted_exp.vp3"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_vp3(f_pattern, file2)
        t_pattern = read_vp3(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->vp3: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_dst(self):
        file1 = "convert_dst.exp"
        file2 = "converted_exp.dst"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_dst(f_pattern, file2)
        t_pattern = read_dst(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->dst: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)

    def test_convert_exp_to_gcode(self):
        file1 = "convert_gcode.exp"
        file2 = "converted_exp.gcode"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_gcode(f_pattern, file2)
        t_pattern = read_gcode(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->gcode: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)
        
    def test_convert_exp_to_xxx(self):
        file1 = "convert_xxx.exp"
        file2 = "converted_exp.xxx"
        write_exp(get_big_pattern(), file1)
        f_pattern = read_exp(file1)
        write_xxx(f_pattern, file2)
        t_pattern = read_xxx(file2)

        self.assertIsNotNone(t_pattern)
        self.assertEqual(t_pattern.count_stitch_commands(COLOR_CHANGE), 15)
        self.assertEqual(t_pattern.count_stitch_commands(STITCH), 16 * 5)
        self.position_equals(t_pattern.stitches, 0, -1)
        print("exp->xxx: ", t_pattern.stitches)
        self.addCleanup(os.remove, file1)
        self.addCleanup(os.remove, file2)


def read_10o_sttiches(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        ctrl = byte[0]
        y = -byte[1]
        x = byte[2]
        if ctrl & 0x20 != 0:
            x = -x
        if ctrl & 0x40 != 0:
            y = -y

        if (ctrl & 0b11111) == 0:
            out.stitch(x, y)
            continue
        if (ctrl & 0b11111) == 0x10:
            out.move(x, y)
            continue
        if ctrl == 0x8A:
            # Start.
            continue
        if ctrl == 0x85:
            out.color_change()
            continue
        if ctrl == 0x82:
            out.stop()
            continue
        if ctrl == 0x81:
            out.trim()
            continue
        if ctrl == 0x87:
            break
        break  # Unknown Control
    out.end()


def read(f, out, settings=None):
    read_10o_sttiches(f, out)



from __future__ import print_function

import unittest

from pyembroidery import *
from pattern_for_tests import *


class TestTrims(unittest.TestCase):

    def test_dst_trims(self):
        file0 = "trim.dst"
        pattern = get_fractal_pattern()
        write_dst(pattern, file0)
        loaded_pattern = read_dst(file0)
        self.assertIsNotNone(loaded_pattern)
        self.assertNotEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        self.addCleanup(os.remove, file0)

    def test_dst_trims_fail(self):
        file0 = "trim.dst"
        pattern = get_fractal_pattern()
        write_dst(pattern, file0)
        loaded_pattern = read_dst(file0, {"trim_at": 50})  # Lines beyond 50 jumps get a trim.
        self.assertIsNotNone(loaded_pattern)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        self.addCleanup(os.remove, file0)

    def test_dst_trims_success(self):
        file0 = "trim.dst"
        pattern = get_fractal_pattern()
        write_dst(pattern, file0, {"trim_at": 50})  # We make trim jumps big enough.
        loaded_pattern = read_dst(file0, {"trim_at": 50, "clipping": False})  # Lines beyond 50 jumps get a trim.
        self.assertIsNotNone(loaded_pattern)
        self.assertNotEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        self.addCleanup(os.remove, file0)

    def test_jef_trims(self):
        file0 = "trim.jef"
        pattern = get_fractal_pattern()
        write_jef(pattern, file0)
        loaded_pattern = read_jef(file0)
        self.assertIsNotNone(loaded_pattern)
        self.assertNotEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        self.addCleanup(os.remove, file0)

    def test_jef_trims_off(self):
        file0 = "trim.jef"
        pattern = get_fractal_pattern()
        write_jef(pattern, file0)
        loaded_pattern = read_jef(file0, {"trim_distance": None})
        self.assertEqual(loaded_pattern.count_stitch_commands(JUMP), 15)
        self.assertIsNotNone(loaded_pattern)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        self.addCleanup(os.remove, file0)

    def test_jef_trims_commands(self):
        file0 = "trim.jef"
        pattern = get_fractal_pattern()
        write_jef(pattern, file0, {"trims": True})
        loaded_pattern = read_jef(file0, {"trim_distance": None})
        self.assertEqual(loaded_pattern.count_stitch_commands(JUMP), 15)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        loaded_pattern = read_jef(file0, {"trim_distance": None, "clipping": False})
        self.assertEqual(loaded_pattern.count_stitch_commands(JUMP), 21)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 0)
        loaded_pattern = read_jef(file0, {"trim_distance": None, "clipping": False, "trims": True, })
        self.assertEqual(loaded_pattern.count_stitch_commands(JUMP), 21)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 2)
        loaded_pattern = read_jef(file0, {"trim_distance": None, "trims": True})
        self.assertEqual(loaded_pattern.count_stitch_commands(JUMP), 15)
        self.assertEqual(loaded_pattern.count_stitch_commands(TRIM), 2)
        self.addCleanup(os.remove, file0)



from .ReadHelper import signed8, read_int_32le


def read_jpx_stitches(f, out):
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        if b[0] != 0x80:
            x = signed8(b[0])
            y = -signed8(b[1])
            out.stitch(x, y)
            continue
        ctrl = b[1]
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        x = signed8(b[0])
        y = -signed8(b[1])
        if ctrl == 0x02:
            out.move(x, y)
            continue
        if ctrl == 0x01:  # Colorchange
            out.color_change()
            if x != 0 and y != 0:
                out.move(x, y)
            continue
        if ctrl == 0x10:
            break
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    stitch_start_position = read_int_32le(f)
    f.seek(0x1C, 1)
    colors = read_int_32le(f)
    f.seek(0x18, 1)
    for i in range(0, colors):
        color_index = read_int_32le(f)
        if color_index is None:
            break
        out.add_thread({
            "color": "random",
            "name": "JPX index " + str(color_index)
        })
    f.seek(stitch_start_position, 0)
    read_jpx_stitches(f, out)



from .EmbThread import EmbThread


def get_thread_set():
    return [
        EmbThreadSew(0, 0, 0, "Unknown", "0"),
        EmbThreadSew(0, 0, 0, "Black", "1"),
        EmbThreadSew(255, 255, 255, "White", "2"),
        EmbThreadSew(255, 255, 23, "Sunflower", "3"),
        EmbThreadSew(250, 160, 96, "Hazel", "4"),
        EmbThreadSew(92, 118, 73, "Green Dust", "5"),
        EmbThreadSew(64, 192, 48, "Green", "6"),
        EmbThreadSew(101, 194, 200, "Sky", "7"),
        EmbThreadSew(172, 128, 190, "Purple", "8"),
        EmbThreadSew(245, 188, 203, "Pink", "9"),
        EmbThreadSew(255, 0, 0, "Red", "10"),
        EmbThreadSew(192, 128, 0, "Brown", "11"),
        EmbThreadSew(0, 0, 240, "Blue", "12"),
        EmbThreadSew(228, 195, 93, "Gold", "13"),
        EmbThreadSew(165, 42, 42, "Dark Brown", "14"),
        EmbThreadSew(213, 176, 212, "Pale Violet", "15"),
        EmbThreadSew(252, 242, 148, "Pale Yellow", "16"),
        EmbThreadSew(240, 208, 192, "Pale Pink", "17"),
        EmbThreadSew(255, 192, 0, "Peach", "18"),
        EmbThreadSew(201, 164, 128, "Beige", "19"),
        EmbThreadSew(155, 61, 75, "Wine Red", "20"),
        EmbThreadSew(160, 184, 204, "Pale Sky", "21"),
        EmbThreadSew(127, 194, 28, "Yellow Green", "22"),
        EmbThreadSew(185, 185, 185, "Silver Grey", "23"),
        EmbThreadSew(160, 160, 160, "Grey", "24"),
        EmbThreadSew(152, 214, 189, "Pale Aqua", "25"),
        EmbThreadSew(184, 240, 240, "Baby Blue", "26"),
        EmbThreadSew(54, 139, 160, "Powder Blue", "27"),
        EmbThreadSew(79, 131, 171, "Bright Blue", "28"),
        EmbThreadSew(56, 106, 145, "Slate Blue", "29"),
        EmbThreadSew(0, 32, 107, "Nave Blue", "30"),
        EmbThreadSew(229, 197, 202, "Salmon Pink", "31"),
        EmbThreadSew(249, 103, 107, "Coral", "32"),
        EmbThreadSew(227, 49, 31, "Burnt Orange", "33"),
        EmbThreadSew(226, 161, 136, "Cinnamon", "34"),
        EmbThreadSew(181, 148, 116, "Umber", "35"),
        EmbThreadSew(228, 207, 153, "Blonde", "36"),
        EmbThreadSew(225, 203, 0, "Sunflower", "37"),
        EmbThreadSew(225, 173, 212, "Orchid Pink", "38"),
        EmbThreadSew(195, 0, 126, "Peony Purple", "39"),
        EmbThreadSew(128, 0, 75, "Burgundy", "40"),
        EmbThreadSew(160, 96, 176, "Royal Purple", "41"),
        EmbThreadSew(192, 64, 32, "Cardinal Red", "42"),
        EmbThreadSew(202, 224, 192, "Opal Green", "43"),
        EmbThreadSew(137, 152, 86, "Moss Green", "44"),
        EmbThreadSew(0, 170, 0, "Meadow Green", "45"),
        EmbThreadSew(33, 138, 33, "Dark Green", "46"),
        EmbThreadSew(93, 174, 148, "Aquamarine", "47"),
        EmbThreadSew(76, 191, 143, "Emerald Green", "48"),
        EmbThreadSew(0, 119, 114, "Peacock Green", "49"),
        EmbThreadSew(112, 112, 112, "Dark Grey", "50"),
        EmbThreadSew(242, 255, 255, "Ivory White", "51"),
        EmbThreadSew(177, 88, 24, "Hazel", "52"),
        EmbThreadSew(203, 138, 7, "Toast", "53"),
        EmbThreadSew(247, 146, 123, "Salmon", "54"),
        EmbThreadSew(152, 105, 45, "Cocoa Brown", "55"),
        EmbThreadSew(162, 113, 72, "Sienna", "56"),
        EmbThreadSew(123, 85, 74, "Sepia", "57"),
        EmbThreadSew(79, 57, 70, "Dark Sepia", "58"),
        EmbThreadSew(82, 58, 151, "Violet Blue", "59"),
        EmbThreadSew(0, 0, 160, "Blue Ink", "60"),
        EmbThreadSew(0, 150, 222, "Solar Blue", "61"),
        EmbThreadSew(178, 221, 83, "Green Dust", "62"),
        EmbThreadSew(250, 143, 187, "Crimson", "63"),
        EmbThreadSew(222, 100, 158, "Floral Pink", "64"),
        EmbThreadSew(181, 80, 102, "Wine", "65"),
        EmbThreadSew(94, 87, 71, "Olive Drab", "66"),
        EmbThreadSew(76, 136, 31, "Meadow", "67"),
        EmbThreadSew(228, 220, 121, "Canary Yellow", "68"),
        EmbThreadSew(203, 138, 26, "Toast", "69"),
        EmbThreadSew(198, 170, 66, "Beige", "70"),
        EmbThreadSew(236, 176, 44, "Honey Dew", "71"),
        EmbThreadSew(248, 128, 64, "Tangerine", "72"),
        EmbThreadSew(255, 229, 5, "Ocean Blue", "73"),
        EmbThreadSew(250, 122, 122, "Sepia", "74"),
        EmbThreadSew(107, 224, 0, "Royal Purple", "75"),
        EmbThreadSew(56, 108, 174, "Yellow Ocher", "76"),
        EmbThreadSew(208, 186, 176, "Beige Grey", "77"),
        EmbThreadSew(227, 190, 129, "Bamboo", "78"),
    ]


class EmbThreadSew(EmbThread):
    def __init__(self, red, green, blue, description, catalog_number):
        EmbThread.__init__(self)
        self.set_color(red, green, blue)
        self.description = description
        self.catalog_number = catalog_number
        self.brand = "Sew"
        self.chart = "Sew"



from .ReadHelper import signed8


def stc_stitch_encoding_read(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break
        x = signed8(byte[0])
        y = -signed8(byte[1])
        ctrl = byte[2]

        if ctrl == 0x01:
            out.stitch(x, y)
            continue
        if ctrl == 0x00:
            out.move(x, y)
            continue
        if ctrl == 25:
            break
        else:
            needle = ctrl - 2
            out.needle_change(needle)
    out.end()


def read(f, out, settings=None):
    f.seek(0x28, 1)  # DESIGN: xxxxxx STITCHES: xxxx.
    stc_stitch_encoding_read(f, out)



def new_stitch_encoding_read(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        x = byte[0]
        y = -byte[1]
        ctrl = byte[2]

        if ctrl & 0b01000000 != 0:
            x = -x
        if ctrl & 0b00100000 != 0:
            y = -y
        ctrl &= ~0b11100000
        if ctrl == 0:
            out.stitch(x, y)
            continue
        if ctrl == 0b00010001:
            break
        if ctrl & 0b00000010 != 0:
            out.color_change()
            continue
        if ctrl & 0b00000001 != 0:
            out.move(x, y)
            continue
    out.end()


def read(f, out, settings=None):
    f.seek(2, 1)  # stitchcount.
    new_stitch_encoding_read(f, out)



import math

from .EmbFunctions import *
from .EmbMatrix import EmbMatrix


class Transcoder:
    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        self.max_stitch = settings.get("max_stitch", float('inf'))
        self.max_jump = settings.get("max_jump", float('inf'))
        self.full_jump = settings.get("full_jump", False)
        self.round = settings.get("round", False)
        self.needle_count = settings.get("needle_count", 5)
        self.thread_change_command = settings.get("thread_change_command", COLOR_CHANGE)
        if self.needle_count <= 1 and self.thread_change_command == NEEDLE_SET:
            self.thread_change_command = STOP
        strip_sequins = settings.get("strip_sequins", True)
        # deprecated, use sequin_contingency.
        if strip_sequins:
            self.sequin_contingency = CONTINGENCY_SEQUIN_JUMP
        else:
            self.sequin_contingency = CONTINGENCY_SEQUIN_UTILIZE
        self.sequin_contingency = settings.get("sequin_contingency", self.sequin_contingency)

        self.writes_speeds = settings.get("writes_speeds", True)
        self.explicit_trim = settings.get("explicit_trim", False)

        self.tie_on_contingency = settings.get("tie_on", CONTINGENCY_TIE_ON_NONE)
        if self.tie_on_contingency is True:
            self.tie_on_contingency = CONTINGENCY_TIE_ON_THREE_SMALL
        if self.tie_on_contingency is False:
            self.tie_on_contingency = CONTINGENCY_TIE_ON_NONE

        self.tie_off_contingency = settings.get("tie_off", CONTINGENCY_TIE_OFF_NONE)
        if self.tie_off_contingency is True:
            self.tie_off_contingency = CONTINGENCY_TIE_OFF_THREE_SMALL
        if self.tie_off_contingency is False:
            self.tie_off_contingency = CONTINGENCY_TIE_OFF_NONE

        self.long_stitch_contingency = \
            settings.get("long_stitch_contingency", CONTINGENCY_LONG_STITCH_JUMP_NEEDLE)

        self.matrix = EmbMatrix()
        translate = settings.get("translate", None)
        if translate is not None:
            try:
                self.matrix.post_translate(translate[0], translate[1])
            except IndexError:
                try:
                    self.matrix.post_translate(translate.x, translate.y)
                except AttributeError:
                    pass
        scale = settings.get("scale", None)
        if scale is not None:
            try:
                self.matrix.post_scale(scale[0], scale[1])
            except (IndexError, TypeError):
                try:
                    self.matrix.post_scale(scale.x, scale.y)
                except AttributeError:
                    self.matrix.post_scale(scale, scale)
        rotate = settings.get("rotate", None)
        if rotate is not None:
            self.matrix.post_rotate(rotate)
        self.source_pattern = None
        self.destination_pattern = None
        self.position = 0
        self.order_index = -1
        self.change_sequence = {}
        self.stitch = None
        self.state_trimmed = True
        self.state_sequin_mode = False
        self.state_jumping = False
        self.needle_x = 0
        self.needle_y = 0
        self.high_flags = 0

    def transcode(self, source_pattern, destination_pattern):
        if source_pattern is destination_pattern:
            # both objects are same. Source is copy, destination is cleared.
            source_pattern = destination_pattern.copy()
            destination_pattern.clear()
        self.source_pattern = source_pattern
        self.destination_pattern = destination_pattern
        self.transcode_metadata()
        self.transcode_main()
        return destination_pattern

    def transcode_metadata(self):
        """Transcodes metadata, (just moves)"""
        source = self.source_pattern.extras
        dest = self.destination_pattern.extras
        dest.update(source)

    def get_as_thread_change_sequence_events(self):
        """Generates the sequence change events with their relevant indexes.
        If there is a sewing event prior to the thread sequence event, the first event
        is indexed as 1. If the first event is a discrete event, occurring before
        the sewing starts it's indexed as zero."""
        source = self.source_pattern.stitches
        current_index = 0
        for stitch in source:
            change = decode_embroidery_command(stitch[2])
            command = change[0]
            flags = command & COMMAND_MASK
            if current_index == 0:
                if flags == STITCH or flags == SEW_TO or flags == NEEDLE_AT or flags == SEQUIN_EJECT:
                    current_index = 1
            if flags == SET_CHANGE_SEQUENCE:
                thread = change[1]
                needle = change[2]
                order = change[3]
                yield flags, thread, needle, order, None
            elif flags == NEEDLE_SET or flags == COLOR_CHANGE or flags == COLOR_BREAK:
                change = decode_embroidery_command(command)
                thread = change[1]
                needle = change[2]
                order = change[3]
                yield flags, thread, needle, order, current_index
                current_index += 1

    def build_thread_change_sequence(self):
        """Builds a change sequence to plan out all the color changes for the file."""
        change_sequence = {}
        lookahead_index = 0
        change_sequence[0] = [None, None, None, None]
        for flags, thread, needle, order, current_index in self.get_as_thread_change_sequence_events():
            if flags == SET_CHANGE_SEQUENCE:
                if order is None:
                    try:
                        current = change_sequence[lookahead_index]
                    except KeyError:
                        current = [None, None, None, None]
                        change_sequence[lookahead_index] = current
                    lookahead_index += 1
                else:
                    try:
                        current = change_sequence[order]
                    except KeyError:
                        current = [None, None, None, None]
                        change_sequence[order] = current
            else:
                try:
                    current = change_sequence[current_index]
                except KeyError:
                    current = [None, None, None, None]
                    change_sequence[current_index] = current
                if current_index >= lookahead_index:
                    lookahead_index = current_index + 1
            if flags == COLOR_CHANGE or flags == NEEDLE_SET:
                current[0] = flags
            if thread is not None:
                current[1] = thread
                current[3] = self.source_pattern.get_thread_or_filler(thread)
            if needle is not None:
                current[2] = needle
        # TODO: account for contingency where threadset repeats threads without explicit values set within the commands.

        needle_limit = self.needle_count
        thread_index = 0
        needle_index = 1
        for order, s in change_sequence.items():
            if s[0] is None:
                s[0] = self.thread_change_command
            if s[1] is None:
                s[1] = thread_index
                thread_index += 1
            if s[2] is None:
                s[2] = needle_index
                if s[2] > needle_limit:
                    s[2] = (s[2] - 1) % needle_limit
                    s[2] += 1
                needle_index += 1
            if s[3] is None:
                s[3] = self.source_pattern.get_thread_or_filler(s[1])
        return change_sequence

    def transcode_main(self):
        """Transcodes stitches.
        Converts middle-level commands and potentially incompatible
        commands into a format friendly low level commands."""

        source = self.source_pattern.stitches
        self.state_trimmed = True
        self.needle_x = 0
        self.needle_y = 0
        self.position = 0
        self.order_index = -1
        self.change_sequence = self.build_thread_change_sequence()

        flags = NO_COMMAND
        for self.position, self.stitch in enumerate(source):
            p = self.matrix.point_in_matrix_space(self.stitch)
            x = p[0]
            y = p[1]
            if self.round:
                x = round(x)
                y = round(y)
            flags = self.stitch[2] & COMMAND_MASK
            self.high_flags = self.stitch[2] & FLAGS_MASK

            if flags == NO_COMMAND:
                continue
            elif flags == STITCH:
                if self.state_trimmed:
                    self.declare_not_trimmed()
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    self.tie_on()
                elif self.state_jumping:
                    self.needle_to(x, y)
                    self.state_jumping = False
                else:
                    self.stitch_with_contingency(x, y)
            elif flags == NEEDLE_AT:
                if self.state_trimmed:
                    self.declare_not_trimmed()
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    self.tie_on()
                elif self.state_jumping:
                    self.needle_to(x, y)
                    self.state_jumping = False
                else:
                    self.needle_to(x, y)
            elif flags == SEW_TO:
                if self.state_trimmed:
                    self.declare_not_trimmed()
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    self.tie_on()
                elif self.state_jumping:
                    self.needle_to(x, y)
                    self.state_jumping = False
                else:
                    self.sew_to(x, y)

            # Middle Level Commands.
            elif flags == STITCH_BREAK:
                self.state_jumping = True
            elif flags == FRAME_EJECT:
                self.tie_off_and_trim_if_needed()
                self.jump_to(x, y)
                self.stop_here()
            elif flags == SEQUENCE_BREAK:
                self.tie_off_and_trim_if_needed()
            elif flags == COLOR_BREAK:
                self.color_break()
            elif flags == TIE_OFF:
                self.tie_off()
            elif flags == TIE_ON:
                self.tie_on()

            # Core Commands.
            elif flags == TRIM:
                self.tie_off_and_trim_if_needed()
            elif flags == JUMP:
                if not self.state_jumping:
                    self.jump_to(x, y)
            elif flags == SEQUIN_MODE:
                self.toggle_sequins()
            elif flags == SEQUIN_EJECT:
                if self.state_trimmed:
                    self.declare_not_trimmed()
                    self.jump_to_within_stitchrange(x, y)
                    self.stitch_at(x, y)
                    self.tie_on()
                if not self.state_sequin_mode:
                    self.toggle_sequins()
                self.sequin_at(x, y)
            elif flags == COLOR_CHANGE:
                self.tie_off_trim_color_change()
            elif flags == NEEDLE_SET:
                self.tie_off_trim_color_change()
            elif flags == STOP:
                self.stop_here()
            elif flags == SLOW:
                self.slow_command_here()
            elif flags == FAST:
                self.fast_command_here()
            elif flags == END:
                self.end_here()
                break
            # On-the-fly Settings Commands.
            elif flags == CONTINGENCY_TIE_ON_THREE_SMALL:
                self.tie_on_contingency = CONTINGENCY_TIE_ON_THREE_SMALL
            elif flags == CONTINGENCY_TIE_OFF_THREE_SMALL:
                self.tie_off_contingency = CONTINGENCY_TIE_OFF_THREE_SMALL
            elif flags == CONTINGENCY_TIE_ON_NONE:
                self.tie_on_contingency = CONTINGENCY_TIE_ON_NONE
            elif flags == CONTINGENCY_TIE_OFF_NONE:
                self.tie_off_contingency = CONTINGENCY_TIE_OFF_NONE
            elif flags == OPTION_MAX_JUMP_LENGTH:
                x = self.stitch[0]
                self.max_jump = x
            elif flags == OPTION_MAX_STITCH_LENGTH:
                x = self.stitch[0]
                self.max_stitch = x
            elif flags == OPTION_EXPLICIT_TRIM:
                self.explicit_trim = True
            elif flags == OPTION_IMPLICIT_TRIM:
                self.explicit_trim = False
            elif flags == CONTINGENCY_LONG_STITCH_NONE:
                self.long_stitch_contingency = CONTINGENCY_LONG_STITCH_NONE
            elif flags == CONTINGENCY_LONG_STITCH_JUMP_NEEDLE:
                self.long_stitch_contingency = CONTINGENCY_LONG_STITCH_JUMP_NEEDLE
            elif flags == CONTINGENCY_LONG_STITCH_SEW_TO:
                self.long_stitch_contingency = CONTINGENCY_LONG_STITCH_SEW_TO
            elif flags == CONTINGENCY_SEQUIN_REMOVE:
                if self.state_sequin_mode:  # if sequin_mode, turn it off.
                    self.toggle_sequins()
                self.sequin_contingency = CONTINGENCY_SEQUIN_REMOVE
            elif flags == CONTINGENCY_SEQUIN_STITCH:
                if self.state_sequin_mode:  # if sequin_mode, turn it off.
                    self.toggle_sequins()
                self.sequin_contingency = CONTINGENCY_SEQUIN_STITCH
            elif flags == CONTINGENCY_SEQUIN_JUMP:
                if self.state_sequin_mode:  # if sequin_mode, turn it off.
                    self.toggle_sequins()
                self.sequin_contingency = CONTINGENCY_SEQUIN_JUMP
            elif flags == CONTINGENCY_SEQUIN_UTILIZE:
                self.sequin_contingency = CONTINGENCY_SEQUIN_UTILIZE
            elif flags == MATRIX_TRANSLATE:
                self.matrix.post_translate(self.stitch[0], self.stitch[1])
            elif flags == MATRIX_SCALE_ORIGIN:
                self.matrix.post_scale(self.stitch[0], self.stitch[1])
            elif flags == MATRIX_ROTATE_ORIGIN:
                self.matrix.post_rotate(self.stitch[0])
            elif flags == MATRIX_SCALE:
                self.matrix.inverse()
                q = self.matrix.point_in_matrix_space(self.needle_x, self.needle_y)
                self.matrix.inverse()
                self.matrix.post_scale(self.stitch[0], self.stitch[1], q[0], q[1])
            elif flags == MATRIX_ROTATE:
                self.matrix.inverse()
                q = self.matrix.point_in_matrix_space(self.needle_x, self.needle_y)
                self.matrix.inverse()
                self.matrix.post_rotate(self.stitch[0], q[0], q[1])
            elif flags == MATRIX_RESET:
                self.matrix.reset()
        if flags != END:
            self.end_here()

    def update_needle_position(self, x, y):
        self.needle_x = x
        self.needle_y = y

    def declare_not_trimmed(self):
        if self.order_index == -1:
            self.next_change_sequence()
        self.state_trimmed = False

    def add_thread_change(self, command, thread=None, needle=None, order=None):
        x = self.needle_x
        y = self.needle_y
        cmd = encode_thread_change(command, thread, needle, order)
        self.destination_pattern.stitches.append([x, y, cmd])

    def add(self, flags, x=None, y=None):
        if x is None:
            x = self.needle_x
        if y is None:
            y = self.needle_y
        flags |= self.high_flags
        self.destination_pattern.stitches.append([x, y, flags])

    def lookahead_stitch(self):
        """Looks forward from current position and
         determines if anymore stitching will occur."""
        source = self.source_pattern.stitches
        for pos in range(self.position, len(source)):
            stitch = source[pos]
            flags = stitch[2]
            if flags == STITCH:
                return True
            elif flags == NEEDLE_AT:
                return True
            elif flags == SEW_TO:
                return True
            elif flags == TIE_ON:
                return True
            elif flags == SEQUIN_EJECT:
                return True
            elif flags == END:
                return False
        return False

    def color_break(self):
        """Implements color break. Should add color changes add needed only."""
        if self.order_index < 0:
            return  # We haven't stitched anything, colorbreak happens, before start. Ignore.
        if not self.state_trimmed:
            self.tie_off()
            if self.explicit_trim:
                self.trim_here()
        if not self.lookahead_stitch():
            return  # No more stitching will happen, colorchange unneeded.
        self.next_change_sequence()
        self.state_trimmed = True

    def tie_off_trim_color_change(self):
        if not self.state_trimmed:
            self.tie_off()
            if self.explicit_trim:
                self.trim_here()
        self.next_change_sequence()
        self.state_trimmed = True

    def tie_off_and_trim_if_needed(self):
        if not self.state_trimmed:
            self.tie_off_and_trim()

    def tie_off_and_trim(self):
        self.tie_off()
        self.trim_here()

    def tie_off(self):
        if self.tie_off_contingency == CONTINGENCY_TIE_OFF_THREE_SMALL:
            try:
                b = self.matrix.point_in_matrix_space(
                    self.source_pattern.stitches[self.position - 1]
                )
                flags = b[2]
                if flags == STITCH or flags == NEEDLE_AT or \
                        flags == SEW_TO or flags == SEQUIN_EJECT:
                    self.lock_stitch(self.needle_x, self.needle_y,
                                     b[0], b[1], self.max_stitch)
            except IndexError:
                pass  # must be an island stitch. jump-stitch-jump

    def tie_on(self):
        if self.tie_on_contingency == CONTINGENCY_TIE_ON_THREE_SMALL:
            try:
                b = self.matrix.point_in_matrix_space(
                    self.source_pattern.stitches[self.position + 1]
                )
                flags = b[2]
                if flags == STITCH or flags == NEEDLE_AT or \
                        flags == SEW_TO or flags == SEQUIN_EJECT:
                    self.lock_stitch(self.needle_x, self.needle_y,
                                     b[0], b[1], self.max_stitch)
            except IndexError:
                pass  # must be an island stitch. jump-stitch-jump

    def trim_here(self):
        if self.state_sequin_mode:
            # Can't trim in sequin mode. DST uses jumps to trigger sequin eject and to trim.
            self.toggle_sequins()
        self.add(TRIM)
        self.state_trimmed = True

    def toggle_sequins(self):
        """Sequin mode toggle can be called whenever but will only actually turn on if set
        to utilize mode for the sequin contingency."""
        contingency = self.sequin_contingency
        if contingency == CONTINGENCY_SEQUIN_UTILIZE:
            self.add(SEQUIN_MODE)
            self.state_sequin_mode = not self.state_sequin_mode

    def jump_to_within_stitchrange(self, new_x, new_y):
        """Jumps close enough to stitch a position in x,y
        without violating the length constraints."""
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_jump
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        if self.full_jump:
            if self.needle_x != new_x or self.needle_y != new_y:
                self.jump_at(new_x, new_y)
        # We are currently assuming that max_jump is also max_stitch.
        # Properly it might be the case that some format could require
        # a split constraint here where we would need to jump further
        # so that we could then stitch closer.

    def jump_to(self, new_x, new_y):
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_jump
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        self.jump_at(new_x, new_y)

    def jump_at(self, new_x, new_y):
        if self.state_sequin_mode:
            self.toggle_sequins()  # can't jump with sequin mode on.
        self.add(JUMP, new_x, new_y)
        self.update_needle_position(new_x, new_y)

    def stitch_with_contingency(self, new_x, new_y):
        if self.long_stitch_contingency == CONTINGENCY_LONG_STITCH_SEW_TO:
            self.sew_to(new_x, new_y)
        elif self.long_stitch_contingency == CONTINGENCY_LONG_STITCH_JUMP_NEEDLE:
            self.needle_to(new_x, new_y)
        else:
            self.stitch_at(new_x, new_y)

    def sew_to(self, new_x, new_y):
        """Stitches to a specific location, with the emphasis on sewing.
         Subdivides long stitches into additional stitches.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, STITCH)
        self.stitch_at(new_x, new_y)

    def needle_to(self, new_x, new_y):
        """Insert needle at specific location, emphasis on the needle.
        Uses jumps to avoid needle penetrations where possible.

        The limit here is the max stitch limit or jump threshold.
        If jump threshold is set low, it will insert jumps even
        between stitches it could have technically encoded values for.

        Stitches to the new location, adding jumps if needed.
        """
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, JUMP)
        self.stitch_at(new_x, new_y)

    def stitch_at(self, new_x, new_y):
        """Inserts a stitch at the specific location.
        Should have already been checked for constraints."""
        self.add(STITCH, new_x, new_y)
        self.update_needle_position(new_x, new_y)

    def sequin_at(self, new_x, new_y):
        """Sequin Ejects at position with the proper Sequin Contingency

        If long stitch contingency is required and since JUMP is often
        sequin eject, the contingency for long stitch is always sew to.
        This shouldn't be exceptionally rare."""
        contingency = self.sequin_contingency
        if contingency == CONTINGENCY_SEQUIN_REMOVE:
            # Do not update the needle position or declare untrimmed.
            return
        x0 = self.needle_x
        y0 = self.needle_y
        max_length = self.max_stitch
        self.interpolate_gap_stitches(x0, y0, new_x, new_y, max_length, STITCH)
        if contingency == CONTINGENCY_SEQUIN_UTILIZE:
            self.add(SEQUIN_EJECT, new_x, new_y)
        elif contingency == CONTINGENCY_SEQUIN_JUMP:
            self.add(JUMP, new_x, new_y)
        elif contingency == CONTINGENCY_SEQUIN_STITCH:
            self.add(STITCH, new_x, new_y)
        self.update_needle_position(new_x, new_y)

    def slow_command_here(self):
        if self.writes_speeds:
            self.add(SLOW)

    def fast_command_here(self):
        if self.writes_speeds:
            self.add(FAST)

    def stop_here(self):
        self.add(STOP)
        self.state_trimmed = True

    def end_here(self):
        self.add(END)
        self.state_trimmed = True

    def next_change_sequence(self):
        self.order_index += 1
        change = self.change_sequence[self.order_index]
        threadlist = self.destination_pattern.threadlist
        threadlist.append(change[3])
        if self.thread_change_command == COLOR_CHANGE:
            if self.order_index != 0:
                self.add_thread_change(COLOR_CHANGE, change[1], change[2])
        elif self.thread_change_command == NEEDLE_SET:
            self.add_thread_change(NEEDLE_SET, change[1], change[2])
        elif self.thread_change_command == STOP:
            self.add_thread_change(STOP, change[1], change[2])
        self.state_trimmed = True

    def position_will_exceed_constraint(self, length=None, new_x=None, new_y=None):
        """Check if the stitch is too long before trying to deal with it."""
        if length is None:
            length = self.max_stitch
        if new_x is None or new_y is None:
            p = self.matrix.point_in_matrix_space(self.stitch[0], self.stitch[1])
            new_x = p[0]
            new_y = p[1]
        distance_x = new_x - self.needle_x
        distance_y = new_y - self.needle_y
        return abs(distance_x) > length or abs(distance_y) > length

    def interpolate_gap_stitches(self, x0, y0, x1, y1, max_length, data):
        """Command sequence line to x, y, respecting length as maximum.
        This does not arrive_at, it steps to within striking distance.
        The next step can arrive at (x, y) without violating constraint.
        If these are already in range, this command will do nothing.

        returns the last stitch interpolated by the code.
        """
        transcode = self.destination_pattern.stitches
        distance_x = x1 - x0
        distance_y = y1 - y0
        if abs(distance_x) > max_length or abs(distance_y) > max_length:
            if data == JUMP and self.state_sequin_mode:
                self.toggle_sequins()  # can't jump with sequin mode on.

            # python 2,3 patch of division that could be integer.
            steps_x = math.ceil(abs(distance_x / (max_length * 1.0)))
            steps_y = math.ceil(abs(distance_y / (max_length * 1.0)))
            if steps_x > steps_y:
                steps = steps_x
            else:
                steps = steps_y
            step_size_x = distance_x / steps
            step_size_y = distance_y / steps
            qx = x0
            qy = y0
            for q in range(1, int(steps)):
                # we need the gap stitches only, not start or end stitch.
                qx += step_size_x
                qy += step_size_y
                stitch = [qx, qy, data | self.high_flags]
                transcode.append(stitch)
                self.update_needle_position(stitch[0], stitch[1])

    def lock_stitch(self, x, y, anchor_x, anchor_y, max_length=None):
        """Tie-on, Tie-off. Lock stitch from current location towards
        anchor location.Ends again at lock location. May not exceed
        max_length in the process."""
        if max_length is None:
            max_length = self.max_stitch
        transcode = self.destination_pattern.stitches
        length = distance(x, y, anchor_x, anchor_y)
        if length > max_length:
            p = oriented(x, y, anchor_x, anchor_y, max_length)
            anchor_x = p[0]
            anchor_y = p[1]
        for amount in (.33, .66, .33, 0):
            transcode.append([
                towards(x, anchor_x, amount),
                towards(y, anchor_y, amount),
                STITCH])


def distance_squared(x0, y0, x1, y1):
    """squared of distance between x0,y0 and x1,y1"""
    dx = x1 - x0
    dy = y1 - y0
    dx *= dx
    dy *= dy
    return dx + dy


def distance(x0, y0, x1, y1):
    """distance between x0,y0 and x1,y1"""
    return math.sqrt(distance_squared(x0, y0, x1, y1))


def towards(a, b, amount):
    """amount between [0,1] -> [a,b]"""
    return (amount * (b - a)) + a


def angle_radians(x0, y0, x1, y1):
    """Angle in radians between x0,y0 and x1,y1"""
    return math.atan2(y1 - y0, x1 - x0)


def oriented(x0, y0, x1, y1, r):
    """from x0,y0 in the direction of x1,y1 in the distance of r"""
    radians = angle_radians(x0, y0, x1, y1)
    return x0 + (r * math.cos(radians)), y0 + (r * math.sin(radians))



from .ReadHelper import read_int_16be, read_int_8, signed8, signed16


def read(f, out, settings=None):
    f.seek(0x11E, 1)
    while True:
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        dy = -signed16(read_int_16be(f))
        dx = signed16(read_int_16be(f))
        c = signed8(read_int_8(f))
        dy -= c
        b = bytearray(f.read(2))
        if len(b) != 2:
            break
        out.stitch(dx, dy)



from __future__ import print_function

import unittest

from pyembroidery import *
from pattern_for_tests import *


class TestInterpolate(unittest.TestCase):

    def test_interpolate_color_stop(self):
        """Full circle STOP/color conversion"""
        pattern = get_fractal_pattern()
        pattern.fix_color_count()
        p2 = pattern.copy()
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)
        pattern.interpolate_stop_as_duplicate_color()
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)

        self.assertEqual(pattern, p2)

    def test_interpolate_to_stop(self):
        """Duplicate color to STOP"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "red"
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)

    def test_interpolate_to_stop_multi(self):
        """Multiple duplicate color to multiple STOP"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "red"
        pattern += (100, 0), (0, 100)
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "red"
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 3)

    def test_interpolate_to_stop_interleave(self):
        """interleaved colors do not become STOP"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "blue"
        pattern += (100, 0), (0, 100)
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "blue"
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)

    def test_interpolate_to_stop_mismatch(self):
        """if the threadlist and color_changes are mismatched"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += "red"
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += COLOR_CHANGE
        pattern += (100, 0), (0, 100)
        pattern += COLOR_CHANGE
        pattern += (0, 0), (100, 100)
        pattern += COLOR_CHANGE
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 2)
        self.assertEqual(pattern.count_stitch_commands(COLOR_CHANGE), 1)

    def test_interpolate_stop_no_threads(self):
        """No threads, cannot duplicate."""
        pattern = EmbPattern()
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        pattern += STOP
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 3)
        pattern.interpolate_stop_as_duplicate_color()
        self.assertEqual(pattern.count_stitch_commands(STOP), 3)

    def test_interpolate_to_stop_intermix(self):
        """intermixed colors become intermixed STOP"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += "red"
        pattern += (100, 0), (0, 100)
        pattern += "blue"
        pattern += (0, 0), (100, 100)
        pattern += "blue"
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        pattern.interpolate_duplicate_color_as_stop()
        self.assertEqual(pattern.count_stitch_commands(STOP), 2)

    def test_interpolate_stop_intermix(self):
        """intermixed STOP become intermixed colors"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        pattern += "blue"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 2)
        pattern.interpolate_stop_as_duplicate_color()
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        self.assertEqual(len(pattern.threadlist),4)
        self.assertEqual(pattern.threadlist[0], pattern.threadlist[1])
        self.assertNotEqual(pattern.threadlist[1], pattern.threadlist[2])
        self.assertEqual(pattern.threadlist[2], pattern.threadlist[3])

    def test_interpolate_stop_duplicate(self):
        """multiple STOP become multiple duplicate"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += STOP
        pattern += STOP
        pattern += STOP
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 4)
        pattern.interpolate_stop_as_duplicate_color()
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        self.assertEqual(len(pattern.threadlist), 5)

    def test_interpolate_stop_duplicate_break(self):
        """changing colors, in a stop series works"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern += STOP
        pattern += STOP
        pattern += "blue"
        pattern += STOP
        pattern += STOP
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(STOP), 4)
        pattern.interpolate_stop_as_duplicate_color()
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)
        self.assertEqual(len(pattern.threadlist), 6)
        self.assertEqual(pattern.threadlist[0], pattern.threadlist[1])
        self.assertEqual(pattern.threadlist[1], pattern.threadlist[2])
        self.assertNotEqual(pattern.threadlist[2], pattern.threadlist[3])
        self.assertEqual(pattern.threadlist[3], pattern.threadlist[4])
        self.assertEqual(pattern.threadlist[4], pattern.threadlist[5])

    def test_interpolate_frame_eject(self):
        """FRAME_EJECTs are interpolated"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern.move_abs(200, 0)
        pattern.stop()
        pattern.move_abs(100, 100)
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 0)
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)
        pattern.interpolate_frame_eject()
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 1)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)

    def test_interpolate_frame_eject_multijums(self):
        """FRAME_EJECTs are interpolated"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern.move_abs(100, 0)
        pattern.move_abs(200, 0)
        pattern.stop()
        pattern.move_abs(100, 0)
        pattern.move_abs(101, 0)
        pattern.move_abs(100, 100)
        pattern += (100, 0), (0, 100)
        self.assertEqual(pattern.count_stitch_commands(JUMP), 5)
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 0)
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)
        pattern.interpolate_frame_eject()
        self.assertEqual(pattern.count_stitch_commands(JUMP), 0)
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 1)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)

    def test_interpolate_frame_eject_end(self):
        """FRAME_EJECTs are interpolated"""
        pattern = EmbPattern()
        pattern += "red"
        pattern += (0, 0), (100, 100)
        pattern.move_abs(200, 0)
        pattern.stop()
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 0)
        self.assertEqual(pattern.count_stitch_commands(STOP), 1)
        pattern.interpolate_frame_eject()
        self.assertEqual(pattern.count_stitch_commands(FRAME_EJECT), 1)
        self.assertEqual(pattern.count_stitch_commands(STOP), 0)



from .EmbFunctions import *
from .EmbThread import EmbThread


def decoded_command(command_dict,name):
    split = name.split(' ')
    command = command_dict[split[0]]
    for sp in split[1:]:
        if sp[0] == "n":
            needle = int(sp[1:])
            command |= (needle + 1) << 16
        if sp[0] == "o":
            order = int(sp[1:])
            command |= (order + 1) << 24
        if sp[0] == "t":
            thread = int(sp[1:])
            command |= (thread + 1) << 8
    return command


def read(f, out, settings=None):
    import json
    json_object = json.load(f)
    command_dict = get_command_dictionary()
    stitches = json_object['stitches']
    extras = json_object['extras']
    threadlist = json_object['threadlist']
    for t in threadlist:
        color = t["color"]
        thread = EmbThread(color)
        thread.description = t["description"]
        thread.catalog_number = t["catalog_number"]
        thread.details = t["details"]
        thread.brand = t["brand"]
        thread.chart = t["chart"]
        thread.weight = t["weight"]
        out.add_thread(thread)
    for s in stitches:
        out.stitches.append(
            [
                s[0],
                s[1],
                decoded_command(command_dict, s[2])
            ]
        )
    out.extras.update(extras)


from .EmbThread import EmbThread
from .ReadHelper import read_int_16be, read_int_8, read_int_32be, read_int_24be, read_signed, read_string_8, \
    read_string_16


def read_vp3_string_16(stream):
    # Reads the header strings which are 16le numbers of size followed by
    # utf-16 text
    string_length = read_int_16be(stream)
    return read_string_16(stream, string_length)


def read_vp3_string_8(stream):
    # Reads the body strings which are 16be numbers followed by utf-8 text
    string_length = read_int_16be(stream)
    return read_string_8(stream, string_length)


def skip_vp3_string(stream):
    string_length = read_int_16be(stream)
    stream.seek(string_length, 1)


def signed32(b):
    b &= 0xFFFFFFFF
    if b > 0x7FFFFFFF:
        return - 0x100000000 + b
    else:
        return b


def signed16(b0, b1):
    b0 &= 0xFF
    b1 &= 0xFF
    b = (b0 << 8) | b1
    if b > 0x7FFF:
        return - 0x10000 + b
    else:
        return b


def read(f, out, settings=None):
    b = f.read(6)
    # magic code: %vsm%\0
    skip_vp3_string(f)  # "Produced by     Software Ltd"
    f.seek(7, 1)
    skip_vp3_string(f)  # "" comments and note string.
    f.seek(32, 1)
    center_x = (signed32(read_int_32be(f)) / 100)
    center_y = -(signed32(read_int_32be(f)) / 100)
    f.seek(27, 1)
    skip_vp3_string(f)  # ""
    f.seek(24, 1)
    skip_vp3_string(f)  # "Produced by     Software Ltd"
    count_colors = read_int_16be(f)
    for i in range(0, count_colors):
        vp3_read_colorblock(f, out, center_x, center_y)
        if (i + 1) < count_colors:  # Don't add the color change on the final read.
            out.color_change()
    out.end()


def vp3_read_colorblock(f, out, center_x, center_y):
    bytescheck = f.read(3)  # \x00\x05\x00
    distance_to_next_block_050 = read_int_32be(f)
    block_end_position = distance_to_next_block_050 + f.tell()

    start_position_x = (signed32(read_int_32be(f)) / 100)
    start_position_y = -(signed32(read_int_32be(f)) / 100)
    abs_x = start_position_x + center_x
    abs_y = start_position_y + center_y
    if abs_x != 0 and abs_y != 0:
        out.move_abs(abs_x, abs_y)
    thread = vp3_read_thread(f)
    out.add_thread(thread)
    f.seek(15, 1)
    bytescheck = f.read(3)  # \x0A\xF6\x00
    stitch_byte_length = block_end_position - f.tell()
    stitch_bytes = read_signed(f, stitch_byte_length)
    i = 0
    while i < len(stitch_bytes) - 1:
        x = stitch_bytes[i]
        y = stitch_bytes[i + 1]
        i += 2
        if (x & 0xFF) != 0x80:
            out.stitch(x, y)
            continue
        if y == 0x01:
            x = signed16(stitch_bytes[i], stitch_bytes[i + 1])
            i += 2
            y = signed16(stitch_bytes[i], stitch_bytes[i + 1])
            i += 2
            out.stitch(x, y)
            i += 2
            # Final element is typically 0x80 0x02, this is skipped regardless of its value.
        elif y == 0x02:
            # This is only seen after 80 01 and should have been skipped. Has no known effect.
            pass
        elif y == 0x03:
            out.trim()


def vp3_read_thread(f):
    thread = EmbThread()
    colors = read_int_8(f)
    transition = read_int_8(f)
    for m in range(0, colors):
        thread.color = read_int_24be(f)
        parts = read_int_8(f)
        color_length = read_int_16be(f)
    thread_type = read_int_8(f)
    weight = read_int_8(f)
    thread.catalog_number = read_vp3_string_8(f)
    thread.description = read_vp3_string_8(f)
    thread.brand = read_vp3_string_8(f)
    return thread



MIT_SIZE_CONVERSION_RATIO = 2.0 / 1.0


def read(f, out, settings=None):
    count = 0
    previous_ctrl = -1
    while True:
        count += 1
        byte = bytearray(f.read(2))
        if len(byte) != 2:
            break
        x = byte[0] & 0x1F
        y = - (byte[1] & 0x1F)
        x *= MIT_SIZE_CONVERSION_RATIO
        y *= MIT_SIZE_CONVERSION_RATIO
        if byte[0] & 0b10000000:
            x = -x
        if byte[1] & 0b10000000:
            y = -y
        ctrl = ((byte[0] & 0x60) >> 3) | ((byte[1] & 0x60) >> 5)
        if ctrl == 0b0111:
            out.stitch(x, y)
            previous_ctrl = ctrl
            continue
        elif ctrl == 0b1100:
            out.move(x, y)
        elif ctrl == 0b0100:
            out.stitch(x, y)
        elif ctrl == 0b0101:
            out.stitch(x, y)
        elif ctrl == 0b1000:
            if previous_ctrl == 0b111:
                out.color_change()
        elif ctrl == 0b0000:
            out.end()  # 0 appears at end.
        else:
            out.stitch(x, y)
        previous_ctrl = ctrl
    out.end()



import datetime

from .EmbThread import build_nonrepeat_palette
from .EmbConstant import *
from .EmbThreadJef import get_thread_set
from .WriteHelper import write_string_utf8, write_int_32le, write_int_8

SEQUIN_CONTINGENCY = CONTINGENCY_SEQUIN_JUMP
FULL_JUMP = True
ROUND = True
MAX_JUMP_DISTANCE = 127
MAX_STITCH_DISTANCE = 127

# These are in mm, embroidery units are 1/10 mm
HOOP_110X110 = 0
HOOP_50X50 = 1
HOOP_140X200 = 2
HOOP_126X110 = 3
HOOP_200X200 = 4


def write(pattern, f, settings=None):
    trims = False
    command_count_max = 3

    date_string = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
    if settings is not None:
        trims = settings.get("trims", trims)
        command_count_max = settings.get('trim_at', command_count_max)
        date_string = settings.get("date", date_string)

    pattern.fix_color_count()
    color_count = pattern.count_threads()
    offsets = 0x74 + (color_count * 8)
    write_int_32le(f, offsets)
    write_int_32le(f, 0x14)
    write_string_utf8(f, date_string)
    write_int_8(f, 0)
    write_int_8(f, 0)
    write_int_32le(f, color_count)
    point_count = 1  # 1 command for END statement
    for stitch in pattern.stitches:
        data = stitch[2] & COMMAND_MASK
        if data == STITCH:
            point_count += 1
        elif data == JUMP:
            point_count += 2
        elif data == TRIM:
            if trims:
                point_count += (2 * command_count_max)
        elif data == COLOR_CHANGE:
            point_count += 2
        elif data == END:
            break
    write_int_32le(f, point_count)
    extends = pattern.bounds()
    design_width = int(round(extends[2] - extends[0]))
    design_height = int(round(extends[3] - extends[1]))
    write_int_32le(f, get_jef_hoop_size(design_width, design_height))
    half_width = int(round(design_width / 2))
    half_height = int(round(design_height / 2))

    # distance from center of hoop.
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)
    write_int_32le(f, half_width)
    write_int_32le(f, half_height)

    # distance from default 110 x 110 hoop
    x_hoop_edge = 550 - half_width
    y_hoop_edge = 550 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 50 x 50 hoop
    x_hoop_edge = 250 - half_width
    y_hoop_edge = 250 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from default 140 x 200 hoop
    x_hoop_edge = 700 - half_width
    y_hoop_edge = 1000 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    # distance from custom hoop, but this should be accepted.
    x_hoop_edge = 700 - half_width
    y_hoop_edge = 1000 - half_height
    write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge)

    jef_threads = get_thread_set()

    palette = build_nonrepeat_palette(jef_threads, pattern.threadlist)
    for t in palette:
        write_int_32le(f, t)

    for i in range(0, color_count):
        write_int_32le(f, 0x0D)

    xx = 0
    yy = 0
    for stitch in pattern.stitches:
        x = stitch[0]
        y = stitch[1]
        data = stitch[2] & COMMAND_MASK
        dx = int(round(x - xx))
        dy = int(round(y - yy))
        xx += dx
        yy += dy
        if data == STITCH:
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == COLOR_CHANGE:
            f.write(b'\x80\x01')
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == TRIM:
            if trims:  # command trim.
                f.write(b'\x80\x02\x00\x00' * command_count_max)
            continue
        elif data == JUMP:
            f.write(b'\x80\x02')
            write_int_8(f, dx)
            write_int_8(f, -dy)
            continue
        elif data == END:
            break
    f.write(b'\x80\x10')


def get_jef_hoop_size(width, height):
    if width < 500 and height < 500:
        return HOOP_50X50
    if width < 1260 and height < 1100:
        return HOOP_126X110
    if width < 1400 and height < 2000:
        return HOOP_140X200
    if width < 2000 and height < 2000:
        return HOOP_200X200
    return HOOP_110X110


def write_hoop_edge_distance(f, x_hoop_edge, y_hoop_edge):
    if min(x_hoop_edge, y_hoop_edge) >= 0:
        write_int_32le(f, x_hoop_edge)  # left
        write_int_32le(f, y_hoop_edge)  # top
        write_int_32le(f, x_hoop_edge)  # right
        write_int_32le(f, y_hoop_edge)  # bottom
    else:
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)
        write_int_32le(f, -1)



from .DstReader import dst_read_header


def z_stitch_encoding_read(f, out):
    count = 0
    while True:
        count += 1
        byte = bytearray(f.read(3))
        if len(byte) != 3:
            break

        y = -byte[0]
        x = byte[1]
        ctrl = byte[2]

        if ctrl & 0x40 != 0:
            x = -x
        if ctrl & 0x20 != 0:
            y = -y

        if (ctrl & 0b11111) == 0:
            out.stitch(x, y)
            continue
        if (ctrl & 0b11111) == 1:
            out.move(x, y)
            continue
        if ctrl == 0x82:
            out.stop()
            continue
        if ctrl == 0x9B:
            out.trim()
            continue
        if 0x83 <= ctrl <= 0x9A:
            needle = (ctrl - 0x83) >> 1
            out.needle_change(needle)
            continue
        break  # Uncaught Control
    out.end()


def read(f, out, settings=None):
    dst_read_header(f, out)
    z_stitch_encoding_read(f, out)




def build_unique_palette(thread_palette, threadlist):
    """Turns a threadlist into a unique index list with the thread palette"""
    chart = [None] * len(thread_palette)  # Create a lookup chart.
    for thread in set(threadlist):  # for each unique color, move closest remaining thread to lookup chart.
        index = thread.find_nearest_color_index(thread_palette)
        if index is None:
            break  # No more threads remain in palette
        thread_palette[index] = None  # entries may not be reused.
        chart[index] = thread  # assign the given index to the lookup.

    palette = []
    for thread in threadlist:  # for each thread, return the index.
        palette.append(thread.find_nearest_color_index(chart))
    return palette


def build_palette(thread_palette, threadlist):
    palette = []
    for thread in threadlist:  # for each thread, return the index.
        palette.append(thread.find_nearest_color_index(thread_palette))
    return palette


def build_nonrepeat_palette(thread_palette, threadlist):
    last_index = None
    last_thread = None
    palette = []
    for thread in threadlist:  # for each thread, return the index.
        index = thread.find_nearest_color_index(thread_palette)
        if last_index == index and last_thread != thread:
            repeated_thread = thread_palette[index]
            repeated_index = index
            thread_palette[index] = None
            index = thread.find_nearest_color_index(thread_palette)
            # index will no longer be repeated.
            thread_palette[repeated_index] = repeated_thread
        palette.append(index)
        last_index = index
        last_thread = thread

    return palette


def find_nearest_color_index(find_color, values):
    if isinstance(find_color, EmbThread):
        find_color = find_color.color
    red = (find_color >> 16) & 0xff
    green = (find_color >> 8) & 0xff
    blue = find_color & 0xff
    closest_index = None
    current_closest_value = float("inf")
    for current_index, t in enumerate(values):
        if t is None:
            continue
        dist = color_distance_red_mean(
            red,
            green,
            blue,
            t.get_red(),
            t.get_green(),
            t.get_blue())
        if dist <= current_closest_value:  # <= choose second if they tie.
            current_closest_value = dist
            closest_index = current_index
    return closest_index


def color_rgb(r, g, b):
    return int(((r & 255) << 16) |
               ((g & 255) << 8) |
               (b & 255))


def color_hex(hex_string):
    h = hex_string.lstrip('#')
    size = len(h)
    if size == 6 or size == 8:
        return int(h[:6], 16)
    elif size == 4 or size == 3:
        return int(h[0] + h[0] + h[1] + h[1] + h[2] + h[2], 16)


def color_distance_red_mean(
        r1, g1, b1,
        r2, g2, b2):
    red_mean = int(round((r1 + r2) / 2))
    r = int(r1 - r2)
    g = int(g1 - g2)
    b = int(b1 - b2)
    return (((512 + red_mean) * r * r) >> 8) + 4 * g * g + \
           (((767 - red_mean) * b * b) >> 8)
    # See the very good color distance paper:
    # https://www.compuphase.com/cmetric.htm


class EmbThread:

    def __init__(self, thread=None, description=None, catalog_number=None, details=None, brand=None, chart=None, weight=None):
        self.color = 0x000000
        self.description = description  # type: str
        self.catalog_number = catalog_number  # type: str
        self.details = details  # type: str
        self.brand = brand  # type: str
        self.chart = chart  # type: str
        self.weight = weight  # type: str
        # description, catalog_number, details, brand, chart, weight
        if thread is not None:
            self.set(thread)

    def __repr__(self):
        parts = list()
        parts.append("thread='%s'" % self.hex_color())
        if self.description is not None:
            parts.append("description='%s'" % self.description)
        if self.catalog_number is not None:
            parts.append("catalog_number='%s'" % self.catalog_number)
        if self.details is not None:
            parts.append("details='%s'" % self.details)
        if self.brand is not None:
            parts.append("brand='%s'" % self.brand)
        if self.chart is not None:
            parts.append("chart='%s'" % self.chart)
        if self.weight is not None:
            parts.append("weight='%s'" % self.weight)
        return "EmbThread(%s)" % ", ".join(parts)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __eq__(self, other):
        if other is None:
            return False
        if isinstance(other, int):
            return self.color & 0xFFFFFF == other & 0xFFFFFF
        try:
            if isinstance(other, basestring):
                return self.color & 0xFFFFFF == EmbThread.parse_string_color(other) & 0xFFFFFF
        except NameError:
            if isinstance(other, str):
                return self.color & 0xFFFFFF == EmbThread.parse_string_color(other) & 0xFFFFFF
        if not isinstance(other, EmbThread):
            return False
        if self.color & 0xFFFFFF != other.color & 0xFFFFFF:
            return False
        if self.description != other.description:
            return False
        if self.catalog_number != other.catalog_number:
            return False
        if self.details != other.details:
            return False
        if self.brand != other.brand:
            return False
        if self.chart != other.chart:
            return False
        if self.weight != other.weight:
            return False
        return True

    def __hash__(self):
        return self.color & 0xFFFFFF

    def __str__(self):
        if self.description is None:
            return "EmbThread %s" % self.hex_color()
        else:
            return "EmbThread %s %s" % (self.description, self.hex_color())

    def set_color(self, r, g, b):
        self.color = color_rgb(r, g, b)

    def get_opaque_color(self):
        return 0xFF000000 | self.color

    def get_red(self):
        red = self.color >> 16
        return red & 0xFF

    def get_green(self):
        green = self.color >> 8
        return green & 0xFF

    def get_blue(self):
        blue = self.color
        return blue & 0xFF

    def find_nearest_color_index(self, values):
        return find_nearest_color_index(int(self.color), values)

    def hex_color(self):
        return "#%02x%02x%02x" % (
            self.get_red(), self.get_green(), self.get_blue()
        )

    def set_hex_color(self, hex_string):
        self.color = color_hex(hex_string)

    def set(self, thread):
        if isinstance(thread, EmbThread):
            self.color = thread.color
            self.description = thread.description
            self.catalog_number = thread.catalog_number
            self.details = thread.details
            self.brand = thread.brand
            self.chart = thread.chart
            self.weight = thread.weight
        elif isinstance(thread, int):
            self.color = thread
        elif isinstance(thread, dict):
            if "name" in thread:
                self.description = thread["name"]
            if "description" in thread:
                self.description = thread["description"]
            if "desc" in thread:
                self.description = thread["desc"]
            if "brand" in thread:
                self.brand = thread["brand"]
            if "manufacturer" in thread:
                self.brand = thread["manufacturer"]
            if "color" in thread or "rgb" in thread:
                try:
                    color = thread["color"]
                except KeyError:
                    color = thread["rgb"]
                if isinstance(color, int):
                    self.color = color
                elif isinstance(color, tuple) or isinstance(color, list):
                    self.color = (color[0] & 0xFF) << 16 | \
                                 (color[1] & 0xFF) << 8 | \
                                 (color[2] & 0xFF)
                else:
                    try:
                        if isinstance(color, basestring):
                            self.color = self.parse_string_color(color)
                    except NameError:
                        if isinstance(color, str):
                            self.color = self.parse_string_color(color)
            if "hex" in thread:
                self.set_hex_color(thread["hex"])
            if "id" in thread:
                self.catalog_number = thread["id"]
            if "catalog" in thread:
                self.catalog_number = thread["catalog"]
        else:
            try:
                if isinstance(thread, basestring):
                    self.color = self.parse_string_color(thread)
            except NameError:
                if isinstance(thread, str):
                    self.color = self.parse_string_color(thread)
            try:  # We might be using python 3.
                if isinstance(thread, long):
                    self.color = int(thread)
            except NameError:
                pass

    @staticmethod
    def parse_string_color(color):
        if color == "random":
            import random
            return random.randint(0, 0xFFFFFF)
        if color[0:1] == "#":
            return color_hex(color[1:])
        color_dict = {
            "aliceblue": color_rgb(240, 248, 255),
            "antiquewhite": color_rgb(250, 235, 215),
            "aqua": color_rgb(0, 255, 255),
            "aquamarine": color_rgb(127, 255, 212),
            "azure": color_rgb(240, 255, 255),
            "beige": color_rgb(245, 245, 220),
            "bisque": color_rgb(255, 228, 196),
            "black": color_rgb(0, 0, 0),
            "blanchedalmond": color_rgb(255, 235, 205),
            "blue": color_rgb(0, 0, 255),
            "blueviolet": color_rgb(138, 43, 226),
            "brown": color_rgb(165, 42, 42),
            "burlywood": color_rgb(222, 184, 135),
            "cadetblue": color_rgb(95, 158, 160),
            "chartreuse": color_rgb(127, 255, 0),
            "chocolate": color_rgb(210, 105, 30),
            "coral": color_rgb(255, 127, 80),
            "cornflowerblue": color_rgb(100, 149, 237),
            "cornsilk": color_rgb(255, 248, 220),
            "crimson": color_rgb(220, 20, 60),
            "cyan": color_rgb(0, 255, 255),
            "darkblue": color_rgb(0, 0, 139),
            "darkcyan": color_rgb(0, 139, 139),
            "darkgoldenrod": color_rgb(184, 134, 11),
            "darkgray": color_rgb(169, 169, 169),
            "darkgreen": color_rgb(0, 100, 0),
            "darkgrey": color_rgb(169, 169, 169),
            "darkkhaki": color_rgb(189, 183, 107),
            "darkmagenta": color_rgb(139, 0, 139),
            "darkolivegreen": color_rgb(85, 107, 47),
            "darkorange": color_rgb(255, 140, 0),
            "darkorchid": color_rgb(153, 50, 204),
            "darkred": color_rgb(139, 0, 0),
            "darksalmon": color_rgb(233, 150, 122),
            "darkseagreen": color_rgb(143, 188, 143),
            "darkslateblue": color_rgb(72, 61, 139),
            "darkslategray": color_rgb(47, 79, 79),
            "darkslategrey": color_rgb(47, 79, 79),
            "darkturquoise": color_rgb(0, 206, 209),
            "darkviolet": color_rgb(148, 0, 211),
            "deeppink": color_rgb(255, 20, 147),
            "deepskyblue": color_rgb(0, 191, 255),
            "dimgray": color_rgb(105, 105, 105),
            "dimgrey": color_rgb(105, 105, 105),
            "dodgerblue": color_rgb(30, 144, 255),
            "firebrick": color_rgb(178, 34, 34),
            "floralwhite": color_rgb(255, 250, 240),
            "forestgreen": color_rgb(34, 139, 34),
            "fuchsia": color_rgb(255, 0, 255),
            "gainsboro": color_rgb(220, 220, 220),
            "ghostwhite": color_rgb(248, 248, 255),
            "gold": color_rgb(255, 215, 0),
            "goldenrod": color_rgb(218, 165, 32),
            "gray": color_rgb(128, 128, 128),
            "grey": color_rgb(128, 128, 128),
            "green": color_rgb(0, 128, 0),
            "greenyellow": color_rgb(173, 255, 47),
            "honeydew": color_rgb(240, 255, 240),
            "hotpink": color_rgb(255, 105, 180),
            "indianred": color_rgb(205, 92, 92),
            "indigo": color_rgb(75, 0, 130),
            "ivory": color_rgb(255, 255, 240),
            "khaki": color_rgb(240, 230, 140),
            "lavender": color_rgb(230, 230, 250),
            "lavenderblush": color_rgb(255, 240, 245),
            "lawngreen": color_rgb(124, 252, 0),
            "lemonchiffon": color_rgb(255, 250, 205),
            "lightblue": color_rgb(173, 216, 230),
            "lightcoral": color_rgb(240, 128, 128),
            "lightcyan": color_rgb(224, 255, 255),
            "lightgoldenrodyellow": color_rgb(250, 250, 210),
            "lightgray": color_rgb(211, 211, 211),
            "lightgreen": color_rgb(144, 238, 144),
            "lightgrey": color_rgb(211, 211, 211),
            "lightpink": color_rgb(255, 182, 193),
            "lightsalmon": color_rgb(255, 160, 122),
            "lightseagreen": color_rgb(32, 178, 170),
            "lightskyblue": color_rgb(135, 206, 250),
            "lightslategray": color_rgb(119, 136, 153),
            "lightslategrey": color_rgb(119, 136, 153),
            "lightsteelblue": color_rgb(176, 196, 222),
            "lightyellow": color_rgb(255, 255, 224),
            "lime": color_rgb(0, 255, 0),
            "limegreen": color_rgb(50, 205, 50),
            "linen": color_rgb(250, 240, 230),
            "magenta": color_rgb(255, 0, 255),
            "maroon": color_rgb(128, 0, 0),
            "mediumaquamarine": color_rgb(102, 205, 170),
            "mediumblue": color_rgb(0, 0, 205),
            "mediumorchid": color_rgb(186, 85, 211),
            "mediumpurple": color_rgb(147, 112, 219),
            "mediumseagreen": color_rgb(60, 179, 113),
            "mediumslateblue": color_rgb(123, 104, 238),
            "mediumspringgreen": color_rgb(0, 250, 154),
            "mediumturquoise": color_rgb(72, 209, 204),
            "mediumvioletred": color_rgb(199, 21, 133),
            "midnightblue": color_rgb(25, 25, 112),
            "mintcream": color_rgb(245, 255, 250),
            "mistyrose": color_rgb(255, 228, 225),
            "moccasin": color_rgb(255, 228, 181),
            "navajowhite": color_rgb(255, 222, 173),
            "navy": color_rgb(0, 0, 128),
            "oldlace": color_rgb(253, 245, 230),
            "olive": color_rgb(128, 128, 0),
            "olivedrab": color_rgb(107, 142, 35),
            "orange": color_rgb(255, 165, 0),
            "orangered": color_rgb(255, 69, 0),
            "orchid": color_rgb(218, 112, 214),
            "palegoldenrod": color_rgb(238, 232, 170),
            "palegreen": color_rgb(152, 251, 152),
            "paleturquoise": color_rgb(175, 238, 238),
            "palevioletred": color_rgb(219, 112, 147),
            "papayawhip": color_rgb(255, 239, 213),
            "peachpuff": color_rgb(255, 218, 185),
            "peru": color_rgb(205, 133, 63),
            "pink": color_rgb(255, 192, 203),
            "plum": color_rgb(221, 160, 221),
            "powderblue": color_rgb(176, 224, 230),
            "purple": color_rgb(128, 0, 128),
            "red": color_rgb(255, 0, 0),
            "rosybrown": color_rgb(188, 143, 143),
            "royalblue": color_rgb(65, 105, 225),
            "saddlebrown": color_rgb(139, 69, 19),
            "salmon": color_rgb(250, 128, 114),
            "sandybrown": color_rgb(244, 164, 96),
            "seagreen": color_rgb(46, 139, 87),
            "seashell": color_rgb(255, 245, 238),
            "sienna": color_rgb(160, 82, 45),
            "silver": color_rgb(192, 192, 192),
            "skyblue": color_rgb(135, 206, 235),
            "slateblue": color_rgb(106, 90, 205),
            "slategray": color_rgb(112, 128, 144),
            "slategrey": color_rgb(112, 128, 144),
            "snow": color_rgb(255, 250, 250),
            "springgreen": color_rgb(0, 255, 127),
            "steelblue": color_rgb(70, 130, 180),
            "tan": color_rgb(210, 180, 140),
            "teal": color_rgb(0, 128, 128),
            "thistle": color_rgb(216, 191, 216),
            "tomato": color_rgb(255, 99, 71),
            "turquoise": color_rgb(64, 224, 208),
            "violet": color_rgb(238, 130, 238),
            "wheat": color_rgb(245, 222, 179),
            "white": color_rgb(255, 255, 255),
            "whitesmoke": color_rgb(245, 245, 245),
            "yellow": color_rgb(255, 255, 0),
            "yellowgreen": color_rgb(154, 205, 50)
        }
        return color_dict.get(color.lower(), 0x000000)
        # return color or black.



from __future__ import print_function

import unittest

from pattern_for_tests import *


class TestEmbpattern(unittest.TestCase):

    def position_equals(self, stitches, j, k):
        self.assertEqual(stitches[j][:1], stitches[k][:1])

    def test_thread_reorder(self):
        test_file = "reorder.pes"
        shift = get_shift_pattern()
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, thread=1, order=0))
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, 0, None, 1))
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, 1, None, 0))
        self.assertEqual(0xFFFFFF & shift.threadlist[0].color, 0xFF0000)
        self.assertEqual(0xFFFFFF & shift.threadlist[1].color, 0x0000FF)
        write_pes(shift, test_file, {"pes version": 6})
        read_pattern = read_pes(test_file)
        for thread in read_pattern.threadlist:
            print(0xFFFFFF & thread.color)
        self.assertEqual((0xFFFFFF & read_pattern.threadlist[0].color), 0x0000FF)
        self.assertEqual((0xFFFFFF & read_pattern.threadlist[1].color), 0xFF0000)
        self.addCleanup(os.remove, test_file)

    def test_needle_count_limited_set(self):
        needle_file = "needle-ls.u01"
        shift = get_shift_pattern()
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, None, 6, 0))
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, 4, 6, 7))
        shift.add_command(encode_thread_change(SET_CHANGE_SEQUENCE, None, 3, 0))
        write_u01(shift, needle_file, {"needle_count": 7})
        needle_pattern = read_u01(needle_file)
        self.assertEqual(needle_pattern.count_stitch_commands(NEEDLE_SET), 16)
        first = True
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            print(cmd)
            if first:
                # self.assertEqual(cmd[2], 3)
                first = False
            self.assertLessEqual(cmd[2], 7)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit1(self):
        needle_file = "needle-1.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 1})
        needle_pattern = read_u01(needle_file)
        self.assertEqual(needle_pattern.count_stitch_commands(STOP), 16)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLess(cmd[2], 1)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit2(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 2})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 2)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit3(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 3})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 3)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit4(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 4})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 4)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit5(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 5})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 5)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit6(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 6})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 6)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit7(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 7})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 7)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit8(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 8})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 8)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit9(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 9})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 9)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_needle_count_limit10(self):
        needle_file = "needle.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 10})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 10)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.addCleanup(os.remove, needle_file)

    def test_u01_tie_on(self):
        needle_file = "tie_on.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 10, "tie_on": CONTINGENCY_TIE_ON_THREE_SMALL})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 10)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.assertEqual(needle_pattern.count_stitch_commands(STITCH), 16 * (5 + 4))
        # 5 for the actual stitch pattern. 3 small, and 1 extra tieon, start.
        self.addCleanup(os.remove, needle_file)

    def test_u01_tie_off(self):
        needle_file = "tie_on.u01"
        write_u01(get_shift_pattern(), needle_file, {"needle_count": 10, "tie_off": CONTINGENCY_TIE_OFF_THREE_SMALL})
        needle_pattern = read_u01(needle_file)
        for stitch in needle_pattern.get_match_commands(NEEDLE_SET):
            cmd = decode_embroidery_command(stitch[2])
            self.assertLessEqual(cmd[2], 10)
        self.assertEqual(needle_pattern.count_needle_sets(), 16)
        self.assertEqual(needle_pattern.count_stitch_commands(STITCH), 16 * (5 + 4))
        # 5 for the actual stitch pattern. 3 small, and 1 extra tieoff, end.
        self.addCleanup(os.remove, needle_file)

    def test_write_dst_read_dst_long_jump(self):
        file1 = "file3.dst"
        pattern = EmbPattern()
        pattern.add_block([(0, 0), (0, 200)], "red")

        write_dst(pattern, file1)
        dst_pattern = read_dst(file1)
        self.assertIsNotNone(dst_pattern)
        self.assertEqual(dst_pattern.count_stitch_commands(STITCH), 2)
        self.assertEqual(dst_pattern.stitches[1][1], 100)
        print("dst: ", dst_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_dst_read_dst_random_stitch(self):
        file1 = "fsmall.dst"
        for i in range(0, 12):
            max = (i * 10) + 1
            write_dst(get_random_pattern_small_halfs(), file1,
                      {"long_stitch_contingency": CONTINGENCY_LONG_STITCH_SEW_TO, "max_stitch": max})
            dst_pattern = read_dst(file1)
            xx = 0
            yy = 0
            command = NO_COMMAND
            for stitch in dst_pattern.stitches:
                dx = stitch[0] - xx
                dy = stitch[1] - yy
                xx += dx
                yy += dy
                last_command = command
                command = stitch[2] & COMMAND_MASK
                if command == STITCH and last_command == STITCH:
                    self.assertLessEqual(dx, max)
                    self.assertLessEqual(dy, max)
            self.assertIsNotNone(dst_pattern)
        self.addCleanup(os.remove, file1)

    def test_write_dst_read_dst_long_jump_random_small(self):
        file1 = "file3small.dst"

        for i in range(0, 1000):
            write_dst(get_random_pattern_small_halfs(), file1,
                      {"long_stitch_contingency": CONTINGENCY_LONG_STITCH_SEW_TO})
            dst_pattern = read_dst(file1)
            self.assertIsNotNone(dst_pattern)
        self.addCleanup(os.remove, file1)

    def test_write_dst_read_dst_long_jump_random_large(self):
        file1 = "file3large.dst"
        for i in range(0, 5):
            write_dst(get_random_pattern_large(), file1,
                      {"long_stitch_contingency": CONTINGENCY_LONG_STITCH_SEW_TO})
            dst_pattern = read_dst(file1)
            self.assertIsNotNone(dst_pattern)
        self.addCleanup(os.remove, file1)

    def test_write_dst_read_dst_divide(self):
        file1 = "file3.dst"
        pattern = EmbPattern()
        pattern.add_block([(0, 0), (0, 2)], "red")

        write_dst(pattern, file1, {"scale": 100, "long_stitch_contingency": CONTINGENCY_LONG_STITCH_SEW_TO})
        dst_pattern = read_dst(file1)
        self.assertIsNotNone(dst_pattern)
        self.assertEqual(dst_pattern.count_stitch_commands(STITCH), 3)
        self.assertEqual(dst_pattern.stitches[1][1], 100)
        print("dst: ", dst_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_csv_read_csv_raw(self):
        file1 = "file.csv"
        write_csv(get_simple_pattern(), file1)
        csv_pattern = read_csv(file1)
        self.assertIsNotNone(csv_pattern)
        self.assertEqual(csv_pattern.count_stitch_commands(COLOR_BREAK), 3)
        self.assertEqual(csv_pattern.count_stitch_commands(STITCH), 15)
        self.position_equals(csv_pattern.stitches, 0, -1)
        print("csv: ", csv_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_csv_read_csv_needle(self):
        file1 = "file2.csv"
        write_csv(get_simple_pattern(), "file2.csv", {"thread_change_command": NEEDLE_SET, "encode": True})
        csv_pattern = read_csv(file1)
        self.assertIsNotNone(csv_pattern)
        self.assertEqual(csv_pattern.count_stitch_commands(NEEDLE_SET), 3)
        self.assertEqual(csv_pattern.count_stitch_commands(STITCH), 15)
        print("csv: ", csv_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_csv_read_csv_color(self):
        file1 = "file3.csv"
        write_csv(get_simple_pattern(), "file3.csv", {"thread_change_command": COLOR_CHANGE, "encode": True})
        csv_pattern = read_csv(file1)
        self.assertEqual(csv_pattern.count_stitch_commands(COLOR_CHANGE), 2)
        self.assertEqual(csv_pattern.count_stitch_commands(STITCH), 15)
        self.position_equals(csv_pattern.stitches, 0, -1)
        print("csv: ", csv_pattern.stitches)
        self.addCleanup(os.remove, file1)

    def test_write_csv_read_csv_encoded_command(self):
        file1 = "file-encoded.csv"
        pattern = EmbPattern()
        encoded_command = encode_thread_change(SET_CHANGE_SEQUENCE, 3, 4, 1)
        pattern.add_command(encoded_command)
        write_csv(pattern, file1)
        csv_pattern = read_csv(file1)
        self.assertIsNotNone(csv_pattern)
        print("csv-encoded: ", csv_pattern.stitches)
        self.assertEqual(encoded_command, csv_pattern.stitches[-1][2])
        self.addCleanup(os.remove, file1)

    def test_issue_87(self):
        """
        Initial test raised by issue 87.
        """
        pattern = EmbPattern()
        stitches_1 = [[0, 1], [2, 3]]
        stitches_2 = [[4, 5], [6, 7]]
        pattern.add_block(stitches_1, 0xFF0000)
        pattern.add_block(stitches_2, 0x0000FF)
        blocks = list(pattern.get_as_colorblocks())
        for q in blocks:
            print(q)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(len(blocks[0][0]), 2)  # 0,1 and 2,3
        self.assertEqual(len(blocks[1][0]), 2)  # 4,5 and 6,7

    def test_issue_87_2(self):
        """
        Tests a pattern arbitrarily starting with a color change.
        With two predefined blocks. The blocks should maintain their blockness.
        The color change should isolate 0 stitches, of an unknown color.
        :return:
        """
        pattern = EmbPattern()
        stitches_1 = [[0, 1], [2, 3]]
        stitches_2 = [[4, 5], [6, 7]]

        pattern.color_change()
        pattern.add_thread('random')
        pattern.add_block(stitches_1, 0xFF0000)
        pattern.add_block(stitches_2, 0x0000FF)
        blocks = list(pattern.get_as_colorblocks())
        # for q in blocks:
        #     print(q)
        self.assertEqual(blocks[1][1].color, 0xFF0000)
        self.assertEqual(blocks[2][1].color, 0x0000FF)
        self.assertEqual(len(blocks), 3)
        self.assertEqual(len(blocks[0][0]), 1)
        self.assertEqual(len(blocks[1][0]), 2)
        self.assertEqual(len(blocks[2][0]), 2)

        for block in blocks:
            stitch_block = block[0]
            for stitch in stitch_block:
                self.assertNotEqual(stitch[2], COLOR_BREAK)

        pattern = EmbPattern()
        pattern.add_thread('random')
        pattern.color_change()  # end block 1, empty
        pattern.add_thread(0xFF0000)
        pattern += stitches_1
        pattern.color_change()  # end block 2
        pattern.add_thread(0x0000FF)
        pattern += stitches_2
        blocks = list(pattern.get_as_colorblocks())
        # end block 3, no explicit end.
        # for q in blocks:
        #     print(q)
        self.assertEqual(blocks[0][0][-1][2], COLOR_CHANGE)  # Color change ends the block.
        self.assertEqual(blocks[1][0][-1][2], COLOR_CHANGE)  # Color change ends the block.
        self.assertEqual(blocks[1][1].color, 0xFF0000)
        self.assertEqual(blocks[2][1].color, 0x0000FF)
        self.assertEqual(len(blocks), 3)
        self.assertEqual(len(blocks[0][0]), 1)
        self.assertEqual(len(blocks[1][0]), 3)
        self.assertEqual(len(blocks[2][0]), 2)  # Final color change is part of no block.
        pattern.color_change()  # end block 3
        blocks = list(pattern.get_as_colorblocks())
        self.assertEqual(len(blocks[2][0]), 3) # Final block with colorchange.

    def test_issue_87_3(self):
        """
        Tests a pattern arbitrarily starting with a needle_set.
        With two predefined blocks. The blocks should maintain their blockness.
        The needle set should not contribute a block. Initial needle_set, only
        define a starting needle.
        :return:
        """
        pattern = EmbPattern()
        pattern.needle_change()
        stitches_1 = [[0, 1], [2, 3]]
        stitches_2 = [[4, 5], [6, 7]]
        pattern.add_block(stitches_1, 0xFF0000)
        pattern.add_block(stitches_2, 0x0000FF)
        blocks = list(pattern.get_as_colorblocks())
        # for q in blocks:
        #     print(q)
        self.assertEqual(blocks[0][1], 0xFF0000)
        self.assertEqual(blocks[1][1], 0x0000FF)
        self.assertEqual(len(blocks), 2)
        for block in blocks:
            stitch_block = block[0]
            for stitch in stitch_block:
                self.assertNotEqual(stitch[2], COLOR_BREAK)

        pattern = EmbPattern()

        pattern.needle_change()  # start block 0
        pattern += stitches_1
        pattern.add_thread(EmbThread(0xFF0000))

        pattern.needle_change()  # start block 1
        pattern += stitches_1
        pattern.add_thread(EmbThread(0x0000FF))

        pattern.needle_change()  # start block 2
        pattern.add_thread(EmbThread('random'))

        blocks = list(pattern.get_as_colorblocks())
        for q in blocks:
            print(q)
        # Mask is required here since needle_set automatically appends extended data.
        self.assertEqual(blocks[0][0][0][2] & COMMAND_MASK, NEEDLE_SET)  # Needle_set starts the block.
        self.assertEqual(blocks[1][0][0][2] & COMMAND_MASK, NEEDLE_SET)  # Needle_set starts the block.
        self.assertEqual(blocks[0][1], 0xFF0000)
        self.assertEqual(blocks[1][1], 0x0000FF)
        self.assertEqual(len(blocks), 3)
        self.assertEqual(len(blocks[0][0]), 3)
        self.assertEqual(len(blocks[1][0]), 3)
        self.assertEqual(len(blocks[2][0]), 1)

    def test_issue_87_4(self):
        """
        Tests a pattern arbitrarily starting with a color break.
        With two predefined blocks. The blocks should maintain their blockness.
        And ending with another arbitrary color break. This should give exactly
        2 blocks which were defined as prepended colorbreaks postpended color breaks
        are not to have an impact.
        :return:
        """
        pattern = EmbPattern()
        pattern += COLOR_BREAK
        stitches_1 = [[0, 1], [2, 3]]
        stitches_2 = [[4, 5], [6, 7]]
        pattern.add_block(stitches_1, 0xFF0000)
        pattern.add_block(stitches_2, 0x0000FF)
        pattern += COLOR_BREAK
        blocks = list(pattern.get_as_colorblocks())
        for q in blocks:
            print(q)

        for block in blocks:
            stitch_block = block[0]
            for stitch in stitch_block:
                self.assertNotEqual(stitch[2], COLOR_BREAK)
        self.assertEqual(blocks[0][1], 0xFF0000)
        self.assertEqual(blocks[1][1], 0x0000FF)
        self.assertEqual(len(blocks), 2)
        self.assertEqual(len(blocks[0][0]), 2)
        self.assertEqual(len(blocks[1][0]), 2)



from .EmbThreadPec import get_thread_set
from .PecReader import read_pec_stitches
from .ReadHelper import read_int_8, read_int_32le, read_int_16le


def read(f, out, settings=None):
    # should start #PHB0003
    f.seek(0x71, 0)
    color_count = read_int_16le(f)
    threadset = get_thread_set()
    for i in range(0, color_count):
        out.add_thread(threadset[read_int_8(f) % len(threadset)])

    file_offset = 0x52

    f.seek(0x54, 0)
    file_offset += read_int_32le(f)

    f.seek(file_offset, 0)
    file_offset += read_int_32le(f) + 2

    f.seek(file_offset, 0)
    file_offset += read_int_32le(f)

    f.seek(file_offset + 14, 0)

    color_count2 = read_int_8(f)
    # 10 bytes unknown, PEC extends.
    f.seek(color_count2 + 0x15, 1)

    read_pec_stitches(f, out)
    out.interpolate_duplicate_color_as_stop()



from .ReadHelper import read_int_16be


def read_zxy_stitches(f, out):
    count = 0
    while True:
        count += 1
        b = bytearray(f.read(3))
        if len(b) != 3:
            break
        ctrl = b[0]
        x = b[1]
        y = -b[2]
        if ctrl & 0x08:
            x = -x
        if ctrl & 0x04:
            y = -y
        ctrl &= ~0x0C
        if ctrl == 0:
            out.stitch(x, y)
            continue
        if ctrl & 0x02:
            out.move(x, y)
            continue
        if ctrl & 0x20:
            if b[1] == 0xFF:
                break
            needle = b[2]
            out.needle_change(needle)
    out.end()


def read(f, out, settings=None):
    f.seek(0x01, 0)
    stitch_start_distance = read_int_16be(f)
    f.seek(stitch_start_distance, 1)
    read_zxy_stitches(f, out)
